@@ -29,7 +29,7 @@
         sports2d --video_input webcam
     - Run with custom parameters (all non specified are set to default): 
         sports2d --show_plots False --time_range 0 2.1 --result_dir path_to_result_dir
-        sports2d --multiperson false --mode lightweight --det_frequency 50
+        sports2d --person_detection_method highest_likelihood --mode lightweight --det_frequency 50
     - Run with a toml configuration file: 
         sports2d --config path_to_config.toml
    
@@ -109,6 +109,7 @@
 
 
 ## INIT
+from importlib.metadata import version
 import argparse
 import toml
 from datetime import datetime
@@ -121,16 +122,17 @@ from Sports2D import Sports2D
 
 
 ## CONSTANTS
-DEFAULT_CONFIG =   {'project': {'video_input': ['demo.mp4'],
-                                'person_height': 1.70,
-                                'load_trc': '',
+DEFAULT_CONFIG =   {'base': {'video_input': ['demo.mp4'],
+                                'nb_persons_to_detect': 'all',
+                                'person_ordering_method': 'on_click',
+                                'first_person_height': 1.65,
+                                'visible_side': ['auto', 'front', 'none'],
+                                'load_trc_px': '',
                                 'compare': False,
                                 'time_range': [],
                                 'video_dir': '',
                                 'webcam_id': 0,
-                                'input_size': [1280, 720]
-                                },
-                    'process': {'multiperson': True,
+                                'input_size': [1280, 720],
                                 'show_realtime_results': True,
                                 'save_vid': True,
                                 'save_img': True,
@@ -143,16 +145,58 @@ DEFAULT_CONFIG =   {'project': {'video_input': ['demo.mp4'],
                                 'pose_model': 'body_with_feet',
                                 'mode': 'balanced',
                                 'det_frequency': 4,
+                                'device': 'auto',
+                                'backend': 'auto',
                                 'tracking_mode': 'sports2d',
-                                'to_meters': True,
-                                'calib_file': '',
-                                'calib_on_person_id': 0,
-                                'floor_angle': 'auto',
-                                'xy_origin': ['auto'],
-                                'save_calib': True,
+                                'deepsort_params': """{'max_age':30, 'n_init':3, 'nms_max_overlap':0.8, 'max_cosine_distance':0.3, 'nn_budget':200, 'max_iou_distance':0.8, 'embedder_gpu': True, 'embedder':'torchreid'}""",
                                 'keypoint_likelihood_threshold': 0.3,
                                 'average_likelihood_threshold': 0.5,
-                                'keypoint_number_threshold': 0.3
+                                'keypoint_number_threshold': 0.3,
+                                'CUSTOM': { 'name': 'Hip',
+                                            'id': 19,
+                                            'children': [{'name': 'RHip',
+                                                'id': 12,
+                                                'children': [{'name': 'RKnee',
+                                                'id': 14,
+                                                'children': [{'name': 'RAnkle',
+                                                    'id': 16,
+                                                    'children': [{'name': 'RBigToe',
+                                                    'id': 21,
+                                                    'children': [{'name': 'RSmallToe', 'id': 23}]},
+                                                    {'name': 'RHeel', 'id': 25}]}]}]},
+                                                {'name': 'LHip',
+                                                'id': 11,
+                                                'children': [{'name': 'LKnee',
+                                                'id': 13,
+                                                'children': [{'name': 'LAnkle',
+                                                    'id': 15,
+                                                    'children': [{'name': 'LBigToe',
+                                                    'id': 20,
+                                                    'children': [{'name': 'LSmallToe', 'id': 22}]},
+                                                    {'name': 'LHeel', 'id': 24}]}]}]},
+                                                {'name': 'Neck',
+                                                'id': 18,
+                                                'children': [{'name': 'Head',
+                                                'id': 17,
+                                                'children': [{'name': 'Nose', 'id': 0}]},
+                                                {'name': 'RShoulder',
+                                                'id': 6,
+                                                'children': [{'name': 'RElbow',
+                                                    'id': 8,
+                                                    'children': [{'name': 'RWrist', 'id': 10}]}]},
+                                                {'name': 'LShoulder',
+                                                'id': 5,
+                                                'children': [{'name': 'LElbow',
+                                                    'id': 7,
+                                                    'children': [{'name': 'LWrist', 'id': 9}]}]}]}]}
+                                },
+                    'px_to_meters_conversion': {
+                                'to_meters': True,
+                                'make_c3d': True,
+                                'calib_file': '',
+                                'floor_angle': 'auto',
+                                'xy_origin': ['auto'],
+                                'save_calib': True
                                 },
                     'angles':   {'display_angle_values_on': ['body', 'list'],
                                 'fontSize': 0.3,
@@ -165,7 +209,9 @@ DEFAULT_CONFIG =   {'project': {'video_input': ['demo.mp4'],
                                                     'Right shoulder',
                                                     'Left shoulder',
                                                     'Right elbow',
-                                                    'Left elbow'],
+                                                    'Left elbow',
+                                                    'Right wrist',
+                                                    'Left wrist'],
                                 'segment_angles': [ 'Right foot',
                                                     'Left foot',
                                                     'Right shank',
@@ -180,7 +226,8 @@ DEFAULT_CONFIG =   {'project': {'video_input': ['demo.mp4'],
                                                     'Left arm',
                                                     'Right forearm',
                                                     'Left forearm'],
-                                'flip_left_right': True
+                                'flip_left_right': True,
+                                'correct_segment_angles_with_floor_angle': True
                                 },
                     'post-processing': {'interpolate': True,
                                         'interp_gap_smaller_than': 10,
@@ -188,21 +235,37 @@ DEFAULT_CONFIG =   {'project': {'video_input': ['demo.mp4'],
                                         'filter': True,
                                         'show_graphs': True,
                                         'filter_type': 'butterworth',
-                                        'butterworth': {'order': 4, 'cut_off_frequency': 3},
+                                        'butterworth': {'order': 4, 'cut_off_frequency': 6},
                                         'gaussian': {'sigma_kernel': 1},
                                         'loess': {'nb_values_used': 5},
                                         'median': {'kernel_size': 3}
                                         },
-                    'inverse-kinematics':{'do_ik': False,
-                                          'person_orientation': ['front', '', 'left'],
+                    'kinematics':{'do_ik': False,
+                                          'use_augmentation': False,
+                                          'feet_on_floor': False,
+                                          'use_simple_model': False,
+                                          'participant_mass': [55.0, 67.0], 
+                                          'right_left_symmetry': True,
+                                          'default_height': 1.70,
+                                          'remove_individual_scaling_setup': True,
+                                          'remove_individual_ik_setup': True,
+                                          'fastest_frames_to_remove_percent': 0.1,
+                                          'close_to_zero_speed_px': 50,
+                                          'close_to_zero_speed_m': 0.2,
+                                          'large_hip_knee_angles': 45,
+                                          'trimmed_extrema_percent': 0.5,
                                           'osim_setup_path': '../OpenSim_setup'
-                                          }
+                                          },
+                    'logging': {'use_custom_logging': False}
                     }
 
 CONFIG_HELP =   {'config': ["C", "path to a toml configuration file"],
                 'video_input': ["i", "webcam, or video_path.mp4, or video1_path.avi video2_path.mp4 ... Beware that images won't be saved if paths contain non ASCII characters"],
-                'person_height': ["H", "height of the person in meters. 1.70 if not specified"],
-                'load_trc': ["", "load trc file to avaid running pose estimation again. false if not specified"],
+                'nb_persons_to_detect': ["n", "number of persons to detect. int or 'all'. 'all' if not specified"],
+                'person_ordering_method': ["", "'on_click', 'highest_likelihood', 'largest_size', 'smallest_size', 'greatest_displacement', 'least_displacement', 'first_detected', or 'last_detected'. 'on_click' if not specified"],
+                'first_person_height': ["H", "height of the reference person in meters. 1.65 if not specified. Not used if a calibration file is provided"],
+                'visible_side': ["", "front, back, left, right, auto, or none. 'auto front none' if not specified. If 'auto', will be either left or right depending on the direction of the motion. If 'none', no IK for this person"],
+                'load_trc_px': ["", "load trc file to avaid running pose estimation again. false if not specified"],
                 'compare': ["", "visually compare motion with trc file. false if not specified"],
                 'webcam_id': ["w", "webcam ID. 0 if not specified"],
                 'time_range': ["t", "start_time end_time. In seconds. Whole video if not specified. start_time1 end_time1 start_time2 end_time2 ... if multiple videos with different time ranges"],
@@ -219,26 +282,39 @@ CONFIG_HELP =   {'config': ["C", "path to a toml configuration file"],
                 'calculate_angles': ["c", "calculate joint and segment angles. true if not specified"],
                 'save_angles': ["A", "save angles as mot files. true if not specified"],
                 'slowmo_factor': ["", "slow-motion factor. For a video recorded at 240 fps and exported to 30 fps, it would be 240/30 = 8. 1 if not specified"],
-                'pose_model': ["p", "only body_with_feet is available for now. body_with_feet if not specified"],
-                'mode': ["m", "light, balanced, or performance. balanced if not specified"],
+                'pose_model': ["p", "body_with_feet, whole_body_wrist, whole_body, or body. body_with_feet if not specified"],
+                'mode': ["m", 'light, balanced, performance, or a """{dictionary within triple quote}""". balanced if not specified. Use a dictionary to specify your own detection and/or pose estimation models (more about in the documentation).'],
                 'det_frequency': ["f", "run person detection only every N frames, and inbetween track previously detected bounding boxes. keypoint detection is still run on all frames.\n\
                                  Equal to or greater than 1, can be as high as you want in simple uncrowded cases. Much faster, but might be less accurate. 1 if not specified: detection runs on all frames"],
+                'backend': ["", "Backend for pose estimation can be 'auto', 'cpu', 'cuda', 'mps' (for MacOS), or 'rocm' (for AMD GPUs)"],
+                'device': ["", "Device for pose estimatino can be 'auto', 'openvino', 'onnxruntime', 'opencv'"],
                 'to_meters': ["M", "convert pixels to meters. true if not specified"],
-                'calib_on_person_id': ["", "person ID to calibrate on. 0 if not specified"],
-                'floor_angle': ["", "angle of the floor. 'auto' if not specified"],
+                'make_c3d': ["", "Convert trc to c3d file. true if not specified"],
+                'floor_angle': ["", "angle of the floor (degrees). 'auto' if not specified"],
                 'xy_origin': ["", "origin of the xy plane. 'auto' if not specified"],
                 'calib_file': ["", "path to calibration file. '' if not specified, eg no calibration file"],
                 'save_calib': ["", "save calibration file. true if not specified"],
                 'do_ik': ["", "do inverse kinematics. false if not specified"],
-                'osim_setup_path': ["", "path to OpenSim setup. '../OpenSim_setup' if not specified"],
-                'person_orientation': ["", "front, back, left, right, auto, or none. 'front none left' if not specified. If 'auto', will be either left or right depending on the direction of the motion."],
-                'multiperson': ["", "multiperson involves tracking: will be faster if set to false. true if not specified"],                'tracking_mode': ["", "sports2d or rtmlib. sports2d is generally much more accurate and comparable in speed. sports2d if not specified"],
+                'use_augmentation': ["", "Use LSTM marker augmentation. false if not specified"],
+                'feet_on_floor': ["", "offset marker augmentation results so that feet are at floor level. true if not specified"],
+                'use_simple_model': ["", ">10 times faster IK if true. No muscles, no constraints (eg stiff spine and shoulders, no patella). false if not specified"],
+                'participant_mass': ["", "mass of the participant in kg or none. Defaults to 70 if not provided. No influence on kinematics (motion), only on kinetics (forces)"],
+                'close_to_zero_speed_m': ["","Sum for all keypoints: about 50 px/frame or 0.2 m/frame"], 
+                'tracking_mode': ["", "'sports2d' or 'deepsort'. 'deepsort' is slower, harder to parametrize but can be more robust if correctly tuned"],
+                'deepsort_params': ["", 'Deepsort tracking parameters: """{dictionary between 3 double quotes}""". \n\
+                                    Default: max_age:30, n_init:3, nms_max_overlap:0.8, max_cosine_distance:0.3, nn_budget:200, max_iou_distance:0.8, embedder_gpu: True\n\
+                                    More information there: https://github.com/levan92/deep_sort_realtime/blob/master/deep_sort_realtime/deepsort_tracker.py#L51'],
                 'input_size': ["", "width, height. 1280, 720 if not specified. Lower resolution will be faster but less precise"],
                 'keypoint_likelihood_threshold': ["", "detected keypoints are not retained if likelihood is below this threshold. 0.3 if not specified"],
                 'average_likelihood_threshold': ["", "detected persons are not retained if average keypoint likelihood is below this threshold. 0.5 if not specified"],
                 'keypoint_number_threshold': ["", "detected persons are not retained if number of detected keypoints is below this threshold. 0.3 if not specified, i.e., i.e., 30 percent"],
+                'fastest_frames_to_remove_percent': ["", "Frames with high speed are considered as outliers. Defaults to 0.1"],
+                'close_to_zero_speed_px': ["", "Sum for all keypoints: about 50 px/frame or 0.2 m/frame. Defaults to 50"],
+                'large_hip_knee_angles': ["", "Hip and knee angles below this value are considered as imprecise. Defaults to 45"],
+                'trimmed_extrema_percent': ["", "Proportion of the most extreme segment values to remove before calculating their mean. Defaults to 50"],
                 'fontSize': ["", "font size for angle values. 0.3 if not specified"],
-                'flip_left_right': ["", "true or false. true to get consistent angles with people facing both left and right sides. Set it to false if you want timeseries to be continuous even when the participent switches their stance. true if not specified"],
+                'flip_left_right': ["", "true or false. Flips angles when the person faces the other side. The person looks to the right if their toe keypoint is to the right of their heel. Set it to false if the person is sprinting or if you want timeseries to be continuous even when the participent switches their stance. true if not specified"],
+                'correct_segment_angles_with_floor_angle': ["", "true or false. If the camera is tilted, corrects segment angles as regards to the floor angle. Set to false is the floor is tilted instead. True if not specified"],
                 'interpolate': ["", "interpolate missing data. true if not specified"],
                 'interp_gap_smaller_than': ["", "interpolate sequences of missing data if they are less than N frames long. 10 if not specified"],
                 'fill_large_gaps_with': ["", "last_value, nan, or zeros. last_value if not specified"],
@@ -248,7 +324,18 @@ CONFIG_HELP =   {'config': ["C", "path to a toml configuration file"],
                 'cut_off_frequency': ["", "cut-off frequency of the Butterworth filter. 3 if not specified"],
                 'sigma_kernel': ["", "sigma of the gaussian filter. 1 if not specified"],
                 'nb_values_used': ["", "number of values used for the loess filter. 5 if not specified"],
-                'kernel_size': ["", "kernel size of the median filter. 3 if not specified"]
+                'kernel_size': ["", "kernel size of the median filter. 3 if not specified"],
+                'osim_setup_path': ["", "path to OpenSim setup. '../OpenSim_setup' if not specified"],
+                'right_left_symmetry': ["", "right left symmetry. true if not specified"],
+                'default_height': ["", "default height for scaling. 1.70 if not specified"],
+                'remove_individual_scaling_setup': ["", "remove individual scaling setup files generated during scaling. true if not specified"],
+                'remove_individual_ik_setup': ["", "remove individual IK setup files generated during IK. true if not specified"],
+                'fastest_frames_to_remove_percent': ["", "Frames with high speed are considered as outliers. Defaults to 0.1"],
+                'close_to_zero_speed_m': ["","Sum for all keypoints: about 0.2 m/frame. Defaults to 0.2"],
+                'close_to_zero_speed_px': ["", "Sum for all keypoints: about 50 px/frame. Defaults to 50"],
+                'large_hip_knee_angles': ["", "Hip and knee angles below this value are considered as imprecise and ignored. Defaults to 45"],
+                'trimmed_extrema_percent': ["", "Proportion of the most extreme segment values to remove before calculating their mean. Defaults to 50"],
+                'use_custom_logging': ["", "use custom logging. false if not specified"]
                 }
 
 
@@ -257,7 +344,7 @@ __author__ = "David Pagnon"
 __copyright__ = "Copyright 2023, Sports2D"
 __credits__ = ["David Pagnon"]
 __license__ = "BSD 3-Clause License"
-__version__ = "0.4.0"
+__version__ = version("sports2d")
 __maintainer__ = "David Pagnon"
 __email__ = "contact@david-pagnon.com"
 __status__ = "Development"
@@ -279,16 +366,16 @@ def base_params(config_dict):
     '''
 
     # video_dir and result_dir
-    video_dir = config_dict.get('project').get('video_dir')
+    video_dir = config_dict.get('base').get('video_dir')
     if video_dir == '': video_dir = Path.cwd()
     else: video_dir = Path(video_dir).resolve()
 
-    result_dir = config_dict.get('process').get('result_dir')
+    result_dir = config_dict.get('base').get('result_dir')
     if result_dir == '': result_dir = Path.cwd()
     else: result_dir = Path(result_dir).resolve()
 
     # video_files, frame_rates, time_ranges
-    video_input = config_dict.get('project').get('video_input')
+    video_input = config_dict.get('base').get('video_input')
     if video_input == "webcam" or video_input == ["webcam"]:
         video_files = ['webcam']  # No video files for webcam
         frame_rates = [None]  # No frame rate for webcam
@@ -306,7 +393,7 @@ def base_params(config_dict):
             video = cv2.VideoCapture(str(video_dir / video_file)) if video_dir else cv2.VideoCapture(str(video_file))
             if not video.isOpened():
                 raise FileNotFoundError(f'Error: Could not open {video_dir/video_file}. Check that the file exists.')
-            frame_rate = video.get(cv2.CAP_PROP_FPS)
+            frame_rate = round(video.get(cv2.CAP_PROP_FPS))
             if frame_rate == 0:
                 frame_rate = 30
                 logging.warning(f'Error: Could not retrieve frame rate from {video_dir/video_file}. Defaulting to 30fps.')
@@ -314,7 +401,7 @@ def base_params(config_dict):
             video.release()
 
         # time_ranges
-        time_ranges = np.array(config_dict.get('project').get('time_range'))
+        time_ranges = np.array(config_dict.get('base').get('time_range'))
         # No time range provided
         if time_ranges.shape == (0,): 
             time_ranges = [None] * len(video_files)
@@ -349,7 +436,7 @@ def get_leaf_keys(config, prefix=''):
 
 def update_nested_dict(config, key_path, value):
     '''
-    Update a nested dictionary based on a key path string like 'process.multiperson'.
+    Update a nested dictionary based on a key path string like 'base.nb_persons_to_detect'.
     '''
 
     keys = key_path.split('.')
@@ -402,11 +489,13 @@ def process(config='Config_demo.toml'):
     else:
         config_dict = read_config_file(config)
     video_dir, video_files, frame_rates, time_ranges, result_dir = base_params(config_dict)
+    use_custom_logging = config_dict.get('logging').get('use_custom_logging')
         
     result_dir.mkdir(parents=True, exist_ok=True)
-    with open(result_dir / 'logs.txt', 'a+') as log_f: pass
-    logging.basicConfig(format='%(message)s', level=logging.INFO, force=True, 
-        handlers = [logging.handlers.TimedRotatingFileHandler(result_dir / 'logs.txt', when='D', interval=7), logging.StreamHandler()]) 
+    if not use_custom_logging: 
+        with open(result_dir / 'logs.txt', 'a+') as log_f: pass
+        logging.basicConfig(format='%(message)s', level=logging.INFO, force=True, 
+            handlers = [logging.handlers.TimedRotatingFileHandler(result_dir / 'logs.txt', when='D', interval=7), logging.StreamHandler()]) 
     
     for video_file, time_range, frame_rate in zip(video_files, time_ranges, frame_rates):
         currentDateAndTime = datetime.now()
@@ -419,7 +508,7 @@ def process(config='Config_demo.toml'):
 
         process_fun(config_dict, video_file, time_range, frame_rate, result_dir)
 
-        elapsed_time = (datetime.now() - currentDateAndTime).total_seconds()        
+        elapsed_time = (datetime.now() - currentDateAndTime).total_seconds()
         logging.info(f'\nProcessing {video_file} took {elapsed_time:.2f} s.')
 
     logging.shutdown()
@@ -445,7 +534,7 @@ def main():
         sports2d --video_input webcam
     - Run with custom parameters (all non specified are set to default): 
         sports2d --show_plots False --time_range 0 2.1 --result_dir path_to_result_dir
-        sports2d --multiperson false --mode lightweight --det_frequency 50
+        sports2d --mode lightweight --det_frequency 50
     - Run with a toml configuration file: 
         sports2d --config path_to_config.toml
     '''
@@ -480,15 +569,16 @@ def main():
     else:
         new_config = DEFAULT_CONFIG.copy()
         if not args.video_input: 
-            new_config.get('project').update({'video_dir': Path(__file__).resolve().parent / 'Demo'})
+            new_config.get('base').update({'video_dir': Path(__file__).resolve().parent / 'Demo'})
 
     # Override dictionary with command-line arguments if provided
     leaf_keys = get_leaf_keys(new_config)
     for leaf_key, default_value in leaf_keys.items():
-        leaf_name = leaf_key.split('.')[-1]
-        cli_value = getattr(args, leaf_name)
-        if cli_value is not None:
-            set_nested_value(new_config, leaf_key, cli_value)
+        if not 'CUSTOM' in leaf_key:
+            leaf_name = leaf_key.split('.')[-1]
+            cli_value = getattr(args, leaf_name)
+            if cli_value is not None:
+                set_nested_value(new_config, leaf_key, cli_value)
 
     # Run process with the new configuration dictionary
     Sports2D.process(new_config)
